<?php

namespace Drupal\jsonapi_taxonomy_tree\Controller;

use Drupal\Core\Cache\CacheableJsonResponse;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Controller\ControllerBase;
use Drupal\taxonomy\VocabularyInterface;
use Symfony\Component\HttpFoundation\JsonResponse;
use GuzzleHttp\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * An example controller.
 */
class JsonapiTaxonomyTree extends ControllerBase {

  /**
   * The HTTP client to fetch the JSON:API data.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  protected $hide_parent = FALSE;

  /**
   * Constructs a JsonapiTaxonomyTree object.
   *
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The HTTP client to fetch the JSON:API data.
   */
  public function __construct(ClientInterface $http_client) {
    $this->httpClient = $http_client;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('http_client')
    );
  }

  /**
   * Makes a request to the JSON:API endpoint.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   *   The JSON response.
   */
  public function fetchJsonapiData(VocabularyInterface $taxonomy_vocabulary) {
    $vid = $taxonomy_vocabulary->id();

    $request = \Drupal::request();
    $queryString = $request->getQueryString();
    $queryParams = [];
    if (!empty($queryString)) {
      parse_str($queryString, $queryParams);
    }

    $this_resource_type = 'taxonomy_term--' . $vid;
    if (isset($queryParams['fields'][$this_resource_type])) {
      $fields_this_resource_type = explode(',', $queryParams['fields'][$this_resource_type]);
      if (!in_array('parent', $fields_this_resource_type)) {
        $this->hide_parent = TRUE;
        $queryParams['fields'][$this_resource_type] = $queryParams['fields'][$this_resource_type] .',parent';
      }
    }

    $queryString = http_build_query($queryParams);
    $baseUrl = \Drupal::service('url_generator')
      ->generateFromRoute('jsonapi.taxonomy_term--' . $vid . '.collection', [], ['absolute' => TRUE]);
    if (!empty($queryString)) {
      $baseUrl = $baseUrl . '?' . $queryString;
    }
    $jsonapi_data = [];
    $nextUrl = $baseUrl;

    $headers = $request->headers->all();
    try {
      do {
        $response = $this->httpClient->get($nextUrl, [
          'headers' => $headers,
        ]);
        $data = json_decode($response->getBody()->getContents(), TRUE);
        $jsonapi_data = array_merge($jsonapi_data, $data['data']);

        // Check if there is a next page
        $nextUrl = isset($data['links']['next']['href']) ? $data['links']['next']['href'] : NULL;
      }
      while ($nextUrl);
      $return_data = [];
      $return_data['cache_time'] = date('Y-m-d H:i:s');
      $return_data['data'] = $this->processData($vid, $jsonapi_data);
      $response = new CacheableJsonResponse($return_data);

      // Handle cache
      $cacheability = (new CacheableMetadata())->addCacheContexts([
        // Make sure that different sparse fieldsets are cached differently.
        'url.query_args:fields',
        // Make sure that different sets of includes are cached differently.
        'url.query_args:include',
        'url.query_args:resourceVersion',
        'url.query_args:filter',
        'url.query_args:sort',
        'url.query_args:page',
      ]);
      $response->addCacheableDependency($cacheability);
      $response->addCacheableDependency((new CacheableMetadata())->addCacheTags(['taxonomy_term_list:'.$vid]));

      return $response;
    }
    catch (\Exception $e) {
      return new JsonResponse(['error' => $e->getMessage()], 500);
    }
  }

  /**
   * Process json api data as taxonomy tree
   *
   * @param string $vid
   * @param $jsonapi_data
   *
   * @return array
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function processData(string $vid, $jsonapi_data) {
    $jsonapi_data = $this->processDataKeyAsUuid($jsonapi_data);

    $data = [];
    /** @var \Drupal\taxonomy\TermStorage $taxonomy_term_storage */
    $taxonomy_term_storage = \Drupal::entityTypeManager()
      ->getStorage('taxonomy_term');
    $first_level_terms = $taxonomy_term_storage->loadTree($vid, 0, 1, TRUE);
    foreach ($first_level_terms as $term) {
      $returned_data = $this->getTermChildren($vid, $term, $taxonomy_term_storage, $jsonapi_data);
      if (!empty($returned_data)) {
        $data[] = $returned_data;
      }
    }
    return $data;
  }

  /**
   * Process json data's key as tid.
   *
   * @param $jsonapi_data
   *
   * @return array
   */
  protected function processDataKeyAsUuid($jsonapi_data) {
    $Processed_jsonapi_data = [];
    foreach ($jsonapi_data as $value) {
      if ($this->hide_parent) {
        unset($value['parent']);
      }
      $Processed_jsonapi_data[$value['id']] = $value;
    }
    return $Processed_jsonapi_data;
  }

  /**
   * Recursively get children of a term.
   *
   * @param string $vid
   * @param $term
   * @param $taxonomy_term_storage
   * @param $jsonapi_data
   *
   * @return array
   */
  private function getTermChildren(string $vid, $term, $taxonomy_term_storage, $jsonapi_data) {
    $term_uuid = $term->uuid();
    if (!isset($jsonapi_data[$term_uuid])) {
      return NULL;
    }
    $term_data = $jsonapi_data[$term_uuid];
    $children = $taxonomy_term_storage->loadTree($vid, $term->id(), 1, TRUE);
    /** @var \Drupal\taxonomy\Entity\Term $child */
    foreach ($children as $child) {
      $returned_data = $this->getTermChildren($vid, $child, $taxonomy_term_storage, $jsonapi_data);
      if (!empty($returned_data)) {
        $term_data['children'][] = $returned_data;
      }
    }
    return $term_data;
  }

}
